;   Export definition file for XPM port to Win32
;   Copyright (C) 2000 Kirill 'Big K' Katsnelson
;
; This file is part of XEmacs.
;
; XEmacs is free software; you can redistribute it and/or modify it
; under the terms of the GNU General Public License as published by the
; Free Software Foundation; either version 2, or (at your option) any
; later version.
;
; XEmacs is distributed in the hope that it will be useful, but WITHOUT
; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
; for more details.
;
; You should have received a copy of the GNU General Public License
; along with XEmacs; see the file COPYING.  If not, write to
; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
; Boston, MA 02111-1307, USA.
;
; Synched up with: Not in FSF.
;

LIBRARY Xpm.dll
EXPORTS
  XpmCreateImageFromData
  XpmCreateDataFromImage
  XpmReadFileToImage
  XpmWriteFileFromImage
  XpmCreateImageFromBuffer
  XpmReadFileToBuffer
  XpmWriteFileFromBuffer
  XpmReadFileToData
  XpmWriteFileFromData
  XpmAttributesSize
  XpmFreeAttributes
  XpmFreeExtensions
  XpmFreeXpmImage
  XpmFreeXpmInfo
  XpmGetErrorString
  XpmLibraryVersion
  XpmReadFileToXpmImage
  XpmWriteFileFromXpmImage
  XpmCreateImageFromXpmImage
  XpmCreateXpmImageFromImage
  XpmCreateDataFromXpmImage
  XpmCreateXpmImageFromData
  XpmCreateXpmImageFromBuffer
  XpmCreateBufferFromXpmImage
  XpmFree
