/* Hey Emacs, this is -*- C -*- code! */

/* Synched up with: Not synched with FSF. */

/* Think twice before editing this file.  Generated automatically by configure.

 The file startup.el guesses at reasonable values for load-path, exec-path,
 and lock-directory.  This means that if you move emacs and its associated
 sub-tree to a different place in the filesystem, or to a different machine,
 you won't have to do anything for it to work.

 If you define the paths in this file then they will take precedence over
 any value generated by the heuristic in startup.el.  The hardcoded paths
 will be checked to see if they are valid, in which case they will be used.
 Otherwise the editor will attempt to make its normal guess.

 See the NEWS file for a description of the heuristic used to locate the lisp
 and exec directories at startup time.  If you are looking at this file
 because you are having trouble, then you would be much better off arranging
 for those heuristics to succeed than defining the paths in this file.

   **  Let me say that again.  If you're editing this file, you're making
   **  a mistake.  Re-read the section on installation in ../etc/NEWS.

 If it defines anything, this file should define some subset of the following:

   PATH_PREFIX		The default value of `prefix-directory'.  This is the
   			default root for everything.

   PATH_LOADSEARCH	The default value of `load-path'.

   PATH_EXEC		The default value of `exec-directory' and `exec-path'.
			(exec-path also contains the value of whatever is in
			the PATH environment variable.)

   PATH_DATA		The default value of `data-directory'.  This
			is where architecture-independent files are
			searched for.

   PATH_LOCK		The name of the directory that contains lock files
			with which we record what files are being modified in 
			Emacs.  This directory should be writable by everyone.
			If this is specified, the string must end with a slash!

   PATH_SUPERLOCK	The name of the file !!!SuperLock!!! in the lock 
			directory.  You probably should let this default...

   PATH_INFO		The default value of `Info-directory-list'.
			This is where info files are searched for. */
