/* The modifiers XEmacs knows about; these appear in key and button events. */

#define XEMACS_MOD_CONTROL	(1<<0)
#define XEMACS_MOD_META		(1<<1)
#define XEMACS_MOD_SUPER	(1<<2)
#define XEMACS_MOD_HYPER	(1<<3)
#define XEMACS_MOD_ALT		(1<<4)
#define XEMACS_MOD_SHIFT	(1<<5)  /* not used for dual-case characters */
#define XEMACS_MOD_BUTTON1	(1<<6)
#define XEMACS_MOD_BUTTON2	(1<<7)
#define XEMACS_MOD_BUTTON3	(1<<8)
#define XEMACS_MOD_BUTTON4	(1<<9)
#define XEMACS_MOD_BUTTON5	(1<<10)
